/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									     *
 * Library:	libsubdiv2.5D						     *
 *									     *
 * Module:	subdiv2.5D_lists.c					     *
 *									     *
 * Description:	lists and trees management functions			     *
 *									     *
 * Date:	20000313						     *
 *									     *
 * Author:	Francisco MORN BURGOS   (http://www.gti.ssr.upm.es./~fmb/)  *
 *		Grupo Tratamiento Imgenes, Univ. Politcnica Madrid, Spain  *
 *									     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



#include <string.h>

#include "subdiv2.5D_priv.h"



/* * *     DETAIL TREES     * * */

sdDetailTree sdlNewLeafDetailTree(char *id)
{
  sdDetailTree dt;

  if (!(dt = NEW(1, sdDetailTreeNode)))
    sdiError("sdlNewLeafDetailTree: not enough memory for a new detail "
	     "tree", -13);
  if (!(dt->id = NEW(strlen(id) + 1, char)))
    sdiError("sdlNewLeafDetailTree: not enough memory for a new detail "
	     "tree ID", -13);
  strcpy(dt->id, id);
  return dt;
} /* sdlNewLeafDetailTree */


#ifdef BIT_SHUFFLING
#  define CODEDELTA_BIGGER(c1, c2)	\
		(c1 << 3 & (CODEWORD)~0) > (c2 << 3 & (CODEWORD)~0)
#else
#  define CODEDELTA_BIGGER(c1, c2)	\
		(c1 & ~MASK_SN) > (c2 & ~MASK_SN)
		/* (c1 << 1 & (CODEWORD)~0) > (c2 << 1 & (CODEWORD)~0) */
#endif

static void sdlCalculateMaxDetailsInAllDTsUnderDT(sdDetailTree dt)
{
  int i;

  if (!dt || !dt->son)
    return;

  for (i=0; i<4; i++)
    sdlCalculateMaxDetailsInAllDTsUnderDT(dt->son[i]);

  dt->D = dt->pt->codedelta;
  for (i=0; i<4; i++) {
    if (CODEDELTA_BIGGER(dt->son[i]->pt->codedelta, dt->D))
      dt->D = dt->son[i]->pt->codedelta;
    if (CODEDELTA_BIGGER(dt->son[i]->D, dt->D))
      dt->D = dt->son[i]->D;
  }

  if (dt->son[0]->son) {
    dt->L = dt->son[0]->D;
    for (i=1; i<4; i++)
      if (CODEDELTA_BIGGER(dt->son[i]->D, dt->L))
	dt->L = dt->son[i]->D;
  }
} /* sdlCalculateMaxDetailsInAllDTsUnderDT */


void sdlCalculateMaxDetailsInAllDetailTrees(sdMeshPtr m)
{
  int i;

  if (!m)
    return;

  for (i=0; i<m->NmotherTriangles; i++)
    sdlCalculateMaxDetailsInAllDTsUnderDT(m->motherDetailTree[i]);
} /* sdlCalculateMaxDetailsInAllDetailTrees */


void sdlDeleteLeafDetailTree(sdDetailTree *dt)
{
  if (!dt || !*dt)
    return;

  DEL((*dt)->id);
  DEL(*dt);
} /* sdlDeleteLeafDetailTree */


static void sdlDestroyAllDTsUnderDT(sdDetailTree *dt)
{
  int i;
  sdDetailTree t;

  if (!dt || !(t = *dt))
    return;

  if (t->son)
    for (i=0; i<4; i++)
      sdlDestroyAllDTsUnderDT(&t->son[i]);
  sdlDeleteLeafDetailTree(dt);
} /* sdlDestroyAllDTsUnderDT */


static void sdlDestroyAllDetailTrees(sdMeshPtr m)
{
  int i;

  if (!m)
    return;

  for (i=0; i<m->NmotherTriangles; i++)
    sdlDestroyAllDTsUnderDT(&m->motherDetailTree[i]);
  DEL(m->motherDetailTree);
} /* sdlDestroyAllDetailTrees */


void sdlAddDetailTree2DTreeList(sdDetailTree dt, int n,
				sdDTreeList *dtl_1st, sdDTreeList *dtl_last)
{
  sdDTreeList dtl_new;

  if (!dt)
    return;

  if (!(dtl_new = NEW(1, sdDTreeListNode)))
    sdiError("sdlAddDetailTree2DTreeList: not enough memory for a new detail "
	     "tree list node", -13);

  dtl_new->dt = dt;
  dtl_new->n = n;
  if (!(dtl_new->prev = *dtl_last))
    *dtl_1st = dtl_new;
  else
    dtl_new->prev->next = dtl_new;
  *dtl_last = dtl_new;
} /* sdlAddDetailTree2DTreeList */


void sdlDeleteDTreeListFromDTreeList(sdDTreeList *dtl, sdDTreeList *dtl_1st,
				     sdDTreeList *dtl_last)
{
  sdDTreeList l;

  if (!dtl || !*dtl || !dtl_1st || !*dtl_1st || !dtl_last || !*dtl_last)
    return;

  l = *dtl;
  if (!l->prev)	/* must be "l = *dtl_1st" */
    *dtl_1st = l->next;
  else
    l->prev->next = l->next;
  if (!l->next)	/* must be "l = *dtl_last" */
    *dtl_last = l->prev;
  else
    l->next->prev = l->prev;
  DEL(*dtl);
} /* sdlDeleteDTreeListFromDTreeList */


void sdlDestroyDTreeList(sdDTreeList *dtl)
{
  sdDTreeList dtlaux;

  if (!dtl)
    return;

  while (dtlaux = *dtl) {
    *dtl = (*dtl)->next;
    DEL(dtlaux);
  }
} /* sdlDestroyDTreeList */



/* * *     POINTS     * * */

#define SON_OF_OT_FOR_V(ot, v)		\
  (v.x < ot->pt->pos.x?			\
    (v.y < ot->pt->pos.y?		\
      (v.z < ot->pt->pos.z? 0 : 1) :	\
      (v.z < ot->pt->pos.z? 2 : 3)) :	\
    (v.y < ot->pt->pos.y?		\
      (v.z < ot->pt->pos.z? 4 : 5) :	\
      (v.z < ot->pt->pos.z? 6 : 7)))

static sdVector sdlPosOff[8] = {{-1.0F, -1.0F, -1.0F}, {-1.0F, -1.0F, +1.0F},
				{-1.0F, +1.0F, -1.0F}, {-1.0F, +1.0F, +1.0F},
				{+1.0F, -1.0F, -1.0F}, {+1.0F, -1.0F, +1.0F},
				{+1.0F, +1.0F, -1.0F}, {+1.0F, +1.0F, +1.0F}};


static void sdlNewLeafPointOT(sdMeshPtr m, sdOctree ot, sdVectorPtr pt)
{
  if (!m || !ot || !pt)
    return;

  if (!(ot->pt = NEW(1, sdPoint)))
    sdiError("sdlNewLeafPointOT: not enough memory for a new point", -13);
  ot->pt->n = m->IDpoint++; m->Npoints++;
  ot->pt->pos = *pt;
} /* sdlNewLeafPointOT */


sdPointPtr sdlAddPoint2VertexOT(sdMeshPtr m, sdOctree *ot, sdVectorPtr v,
				sdBoolean *found)
{
  sdOctree t;
  int i, s, sold;
  sdPointPtr pold;
  sdVector d;
  float pw;

  if (!m || !ot || !v)
    return NULL;

  *found = FALSE;

  if (!*ot) {
    if (!(*ot = NEW(1, sdOctreeNode)))
      sdiError("sdlAddPoint2VertexOT: not enough memory for a new octree", -13);
    m->NnodesOT++;
  }

  t = *ot;
  if (t->tree) {
    s = SON_OF_OT_FOR_V(t, (*v));
    if (!t->son[s]) {
      if (!(t->son[s] = NEW(1, sdOctreeNode)))
	sdiError("sdlAddPoint2VertexOT: not enough memory for a new tree", -13);
      m->NnodesOT++;
      t->son[s]->depth = t->depth + 1;
      t->son[s]->mom = t;
    }
    return sdlAddPoint2VertexOT(m, &t->son[s], v, found);
  }

  if (t->pt) {
    d = *sdmVecDiff(v, &t->pt->pos);
    if (NEGLIGIBLE(d.x) && NEGLIGIBLE(d.y) && NEGLIGIBLE(d.z)) {
      *found = TRUE;
      return t->pt;
    }

    pold = t->pt;
    for (i=0, pw=1.0F; i<t->depth; i++)
      pw *= 0.5F;
    do {
      t->tree = TRUE;
      if (!(t->pt = NEW(1, sdPoint)))
	sdiError("sdlAddPoint2VertexOT: not enough memory for a new point",
		 -13);
      t->pt->n = -1;
      if (t->depth == 0)
	t->pt->pos = sdNullVector;
      else {
	i = SON_OF_OT_FOR_V(t->mom, pold->pos);
	t->pt->pos = sdlPosOff[i];
	MUL_VEC(t->pt->pos, pw);
	INC_VEC(t->pt->pos, t->mom->pt->pos);
      }
      s    = SON_OF_OT_FOR_V(t, (*v));
      sold = SON_OF_OT_FOR_V(t, pold->pos);
      if (!(t->son = NEW(8, sdOctree)) || !(t->son[s] = NEW(1, sdOctreeNode)))
	sdiError("sdlAddPoint2VertexOT: not enough memory for a new tree", -13);
      m->NnodesOT++;
      if (s != sold) {
	if (!(t->son[sold] = NEW(1, sdOctreeNode)))
	  sdiError("sdlAddPoint2VertexOT: not enough memory for a new tree",
		   -13);
	m->NnodesOT++;
      }
      t->son[s]->depth = t->son[sold]->depth = t->depth + 1;
      t->son[s]->mom = t->son[sold]->mom = t;
      t = t->son[sold];
      pw *= 0.5F;
    } while (s == sold);
    t->pt = pold;
    t = t->mom->son[s];
  }

  sdlNewLeafPointOT(m, t, v);
  return t->pt;
} /* sdlAddPoint2VertexOT */


void sdlDeletePointFromVertexOT(sdMeshPtr m, sdOctree *ot, sdPointPtr *pt)
{
  sdOctree t;
  int i, nemptysons, s, saux;

  if (!m || !ot || !*ot || !pt || !*pt)
    return;

  t = *ot;
  if (t->tree) {
    /* 1. recurse */
    s = SON_OF_OT_FOR_V(t, (*pt)->pos);
    if (!t->son[s])
      sdiError("sdlDeletePointFromVertexOT: tried to delete a ghost point",
	       -10);
    sdlDeletePointFromVertexOT(m, &t->son[s], pt);

    /* [2. perform tree-collapsing if needed] */
    for (i=nemptysons=0, saux=-1; i<8; i++) {
      if (!t->son[i])
	nemptysons++;
      else if (!t->son[i]->tree)
	saux = i;
    }
    if (nemptysons == 7 && saux != -1) {
      (*ot) = t->son[saux];
      (*ot)->depth--;
      (*ot)->mom = t->mom;
      DEL(t->son);
      DEL(t->pt);
      DEL(t);
      m->NnodesOT--;
    }
    return;
  }

  if (t->pt != *pt)
    sdiError("sdlDeletePointFromVertexOT: tried to delete a ghost point", -10);
  DEL(t->pt);
  DEL(*ot);
  DEL(*pt);
  m->Npoints--; m->NnodesOT--;
} /* sdlDeletePointFromVertexOT */


static void sdlPrintAllPointsInOT(sdMeshPtr m, sdOctree ot,
				  sdOutputFormat format, sdBoolean normal)
{
  int i, spc;

  if (!m)
    return;

  switch (format) {
    case USELESS:
      if (!ot) {
        printf("*\n");
        return;
      } else if (ot->tree) {	/* virtual vertex: recurse */
        printf("TREE (%d) rooted at ", ot->depth);
        sdiPrintPoint(m, ot->pt, USELESS, normal);
        for (spc=0; spc<ot->depth; spc++)
          printf("   ");
        printf("{\n");
        for (i=0; i<8; i++) {
          for (spc=0; spc<ot->depth; spc++)
            printf("   ");
          printf("%d: ", i);
          sdlPrintAllPointsInOT(m, ot->son[i], USELESS, normal);
        }
        for (spc=0; spc<ot->depth; spc++)
          printf("   ");
        printf("}\n");
      } else {			/* real vertex: print it */
        printf("LEAF (%d): ", ot->depth);
	ot->pt->n = m->IDpoint++;
        sdiPrintPoint(m, ot->pt, USELESS, normal);
      }
      break;

    case MATHEMATICA:
      /* NOP: no vertices list in Mathematica format */
      break;

    case VRML2:
      if (!ot)
        return;
      if (ot->tree) {	/* virtual vertex: recurse */
        for (i=0; i<8; i++)
          if (ot->son[i])
            sdlPrintAllPointsInOT(m, ot->son[i], VRML2, normal);
      } else {		/* real vertex: print it */
	ot->pt->n = m->IDpoint++;
        sdiPrintPoint(m, ot->pt, VRML2, normal);
      }
      break;

    default:
      sdiError("sdlPrintAllPointsInOT: unknown output format", -3);
  }
} /* sdlPrintAllPointsInOT */


void sdlPrintAllPointsInVertexOT(sdMeshPtr m, sdOutputFormat format,
				 sdBoolean normal)
{
  if (!m)
    return;

  m->IDpoint = 0;
  sdlPrintAllPointsInOT(m, m->vertexOT, format, normal);
} /* sdlPrintAllPointsInVertexOT */


static void sdlDestroyOctree(sdMeshPtr m, sdOctree *ot, sdBoolean freePoints)
{
  sdOctree t;
  int i;

  if (!ot || !(t = *ot))
    return;

  if (t->tree) {
    for (i=0; i<8; i++)
      if (t->son[i])
	sdlDestroyOctree(m, &t->son[i], freePoints);
    DEL(t->son);
    DEL(t->pt);
  } else {
    if (freePoints)
      DEL(t->pt);
  }
  DEL(*ot);
  m->NnodesOT--;
} /* sdlDestroyOctree */


static void sdlAddAllPointsInOT2VertexList(sdMeshPtr m, sdOctree ot)
{
  int i;
  sdPointPtr pt;

  if (!ot || !m)
    return;
 
  if (ot->tree)	/* virtual vertex: recurse */
    for (i=0; i<8; i++)
      sdlAddAllPointsInOT2VertexList(m, ot->son[i]);
  else {	/* real vertex: add it to the list */
    pt = ot->pt;
    pt->n = m->IDpoint++; m->Npoints++;
    if (!(pt->prev = m->lastVertex))
      if (!m->firstVertex)
	m->firstVertex = pt;
      else
	sdiError("sdlAddAllPointsInOT2VertexList: yuyu", -8);
    else
      pt->prev->next = pt;
    m->lastVertex = pt;
  }
} /* sdlAddAllPointsInOT2VertexList */


void sdlTransferAllVerticesFromVertexOT2VertexList(sdMeshPtr m)
{
  if (!m)
    return;

  m->IDpoint = m->Npoints = 0;
  sdlAddAllPointsInOT2VertexList(m, m->vertexOT);
  sdlDestroyOctree(m, &m->vertexOT, FALSE);
} /* sdlTransferAllVerticesFromVertexOT2VertexList */


sdPointPtr sdlAddPoint2VertexList(sdMeshPtr m, sdVectorPtr pt)
{
  sdPointPtr pt_new;

  if (!m || !pt)
    return NULL;

  if (!(pt_new = NEW(1, sdPoint)))
    sdiError("sdlAddPoint2VertexList: not enough memory for a new point", -13);

  pt_new->n = m->IDpoint++; m->Npoints++;
  pt_new->pos = *pt;
  if (!(pt_new->prev = m->lastVertex))
    m->firstVertex = pt_new;
  else
    pt_new->prev->next = pt_new;
  m->lastVertex = pt_new;
  return pt_new;
} /* sdlAddPoint2VertexList */


void sdlDeletePointFromVertexList(sdMeshPtr m, sdPointPtr *pt)
{
  if (!m || !pt || !*pt)
    return;

  if (!(*pt)->prev)
    m->firstVertex = (*pt)->next;
  else
    (*pt)->prev->next = (*pt)->next;
  if (!(*pt)->next)
    m->lastVertex = (*pt)->prev;
  else
    (*pt)->next->prev = (*pt)->prev;
  DEL(*pt);
  m->IDpoint--; m->Npoints--;
} /* sdlDeletePointFromVertexList */


static void sdlDestroyVertexList(sdMeshPtr m)
{
  sdPointPtr pt;

  if (!m)
    return;

  while (m->firstVertex) {
    pt = m->firstVertex->next;
    DEL(m->firstVertex);
    m->firstVertex = pt;
  }
  m->lastVertex = NULL;
} /* sdlDestroyVertexList */


sdPointPtr sdlFindVertexByID(sdMeshPtr m, int id)
{
  sdPointPtr pt;

  if (!m)
    return NULL;

  for (pt=m->lastVertex; pt; pt=pt->prev)
    if (pt->n < id) break;
  return (pt->next->n == id? pt->next : NULL);
} /* sdlFindVertexByID */


int sdlVertexDepth(sdMeshPtr m, sdPointPtr pt)
{
  int d;
  sdTrianglePtr t;

  if (!m || !pt)
    return -1;

  for (t=m->motherTriangle[0], d=0; t->subdiv; t=t->son[0], d++)
    if (pt->n < t->son[0]->vtx[1]->n) break;
  return d;
} /* sdlVertexDepth */


typedef struct _VNN *sdlVtxNbrList;
typedef struct _VNN {
  sdPointPtr	nbr;
  sdlVtxNbrList	next;
} sdlVtxNbrNode;


static void sdlDestroyVNL(sdlVtxNbrList *vnl)
{
  sdlVtxNbrList vnlaux;

  if (!vnl)
    return;

  while (*vnl) {
    vnlaux = (*vnl)->next;
    DEL(*vnl);
    *vnl = vnlaux;
  }
} /* sdlDestroyVNL */


static int sdlBuildVNL(sdPointPtr vtx, sdlVtxNbrList *vnl)
{
  int Nnbrs, i;
  sdlVtxNbrList vnl_new;
  sdTrianglePtr t, t_prv;
  sdPointPtr nbr;
  sdVector normal, naux;

  if (!vtx || !vtx->t || !vnl)
    return -1;

  if (*vnl)
    sdlDestroyVNL(vnl);

  /* push all neighbour vertices onto vnl in cw order so that they can	*/
  /* be popped out of it in ccw order in "sdlComputeAllVertexNormals"	*/
  /* and used later by "sdmFillVertexNormal"				*/
  /* NB: if vertex normal is computed as a weighted average of incident	*/
  /* triangle normals (weights = triangle areas), triangle normals and	*/
  /* areas must be computed before vertex normals			*/
  Nnbrs = 0;
  t = vtx->t;
  do {
    i = IDX_VTX_IN_T(vtx, t);
    nbr = t->vtx[sdP1[i]];
    if (!(vnl_new = NEW(1, sdlVtxNbrNode)))
      sdiError("sdlBuildVNL: not enough memory for a new VNN", -13);
    vnl_new->nbr = nbr;
    vnl_new->next = *vnl;
    *vnl = vnl_new;
    Nnbrs++;
    t_prv = t;
  } while ((t = t->adt[i]) && t != vtx->t);

  /* if a border edge has been reached, go back ccw until the other	*/
  /* border edge is reached, computing indeed the vertex normal as the	*/
  /* weighted average of its surrounding triangles; destroy "vnl" and	*/
  /* return "Nnbrs = -1" so that "vnl" is not used by the caller	*/
  if (!t) {
    normal = sdNullVector;
    t = t_prv;
    do {
      naux = t->normal;
      MUL_VEC(naux, t->K);
      INC_VEC(normal, naux);
      i = sdM1[IDX_VTX_IN_T(vtx, t)];
    } while (t = t->adt[i]);
    vtx->normal = *sdmNormalise(&normal);
    sdlDestroyVNL(vnl);
    Nnbrs = -1;
  }

  return Nnbrs;
} /* sdlBuildVNL */


void sdlComputeAllVertexNormals(sdMeshPtr m)
{
  sdPointPtr pt, *nbr;
  sdlVtxNbrList vnl = NULL, vnlaux;
  int n, i;

  if (!m || m->nonManifold)
    return;

  for (pt=m->firstVertex; pt; pt=pt->next)
    if ((n = sdlBuildVNL(pt, &vnl)) > 0) {
      if (!(nbr = NEW(n, sdPointPtr)))
	sdiError("sdlComputeAllVertexNormals: not enough memory for a new VN",
		 -13);
      for (i=0, vnlaux=vnl; i<n; i++, vnlaux=vnlaux->next)
	nbr[i] = vnlaux->nbr;
      sdlDestroyVNL(&vnl);
      sdmFillVertexNormal(pt, n, nbr);
      DEL(nbr);
    }
} /* sdlComputeAllVertexNormals */



/* * *     TRIANGLES     * * */

sdTrianglePtr sdlNewLeafTriangle(sdMeshPtr m, int depth)
{
  sdTrianglePtr t;

  if (!m)
    return NULL;

  if (!(t = NEW(1, sdTriangle)))
    sdiError("sdlNewLeafTriangle: not enough memory for a new triangle", -13);
  t->n = m->IDtriangle++;
  t->depth = depth;
  m->NleafTriangles++; m->Ntriangles++;
  return t;
} /* sdlNewLeafTriangle */


void sdlDeleteLeafTriangle(sdMeshPtr m, sdTrianglePtr *t)
{
  if (!m || !t || !*t)
    return;

  DEL(*t);
  m->NleafTriangles--; m->Ntriangles--;
} /* sdlDeleteLeafTriangle */


static sdTrianglePtr sdlFindLeafTriangleByIDinMT(int id, sdTrianglePtr t)
{
  int i;
  sdTrianglePtr ts;

  if (!t)
    return NULL;

  if (t->subdiv) {
    for (i=0; i<4; i++)
      if (ts = sdlFindLeafTriangleByIDinMT(id, t->son[i]))
	return ts;
    return NULL;
  }

  return (t->n == id? t : NULL);
} /* sdlFindLeafTriangleByIDinMT */


/* ToDo: improve search by id (see also "sdg{Pick,Display}Triangle()") */
sdTrianglePtr sdlFindLeafTriangleByID(sdMeshPtr m, int id)
{
  int i;
  sdTrianglePtr t;

  if (!m)
    return NULL;

  for (i=0; i<m->NmotherTriangles; i++)
    if (t = sdlFindLeafTriangleByIDinMT(id, m->motherTriangle[i]))
      return t;
  return NULL;
} /* sdlFindLeafTriangleByID */


static void sdlDestroyAllTrianglesUnderMT(sdMeshPtr m, sdTrianglePtr *t)
{
  int i;

  if (!m || !t || *t)
    return;

  if ((*t)->subdiv) {
    for (i=0; i<4; i++)
      sdlDestroyAllTrianglesUnderMT(m, &(*t)->son[i]);
    m->NleafTriangles++;
  }

  sdlDeleteLeafTriangle(m, t);
} /* sdlDestroyAllTrianglesUnderMT */


static void sdlDestroyAllTriangles(sdMeshPtr m)
{
  int i;

  if (!m)
    return;

  for (i=0; i<m->NmotherTriangles; i++)
    sdlDestroyAllTrianglesUnderMT(m, &m->motherTriangle[i]);
  DEL(m->motherTriangle);
} /* sdlDestroyAllTriangles */


void sdlAddTriangle2TriangleList(sdTrianglePtr t, sdTriangleLst *tl)
{
  sdTriangleLst tl_crt, tl_new;

  if (!t || !tl)
    return;

  for (tl_crt=*tl; tl_crt; tl_crt=tl_crt->next)
    if (tl_crt->t == t)
      return;

  if (!(tl_new = NEW(1, sdTriangleLstNode)))
    sdiError("sdlAddTriangle2TriangleList: not enough memory for a new "
	     "triangle", -13);

  tl_new->t = t;
  tl_new->next = *tl;
  *tl = tl_new;
} /* sdlAddTriangle2TriangleList */


sdBoolean sdlIsTriangleInTriangleList(sdTrianglePtr t, sdTriangleLst tl)
{
  sdTriangleLst tl_crt;

  if (!t || !tl)
    return FALSE;

  for (tl_crt=tl; tl_crt; tl_crt=tl_crt->next)
    if (tl_crt->t == t)
      return TRUE;

  return FALSE;
} /* sdlIsTriangleInTriangleList */


void sdlDestroyTriangleList(sdTriangleLst *tl)
{
  sdTriangleLst tlaux;

  if (!tl)
    return;

  while (tlaux = *tl) {
    *tl = (*tl)->next;
    DEL(tlaux);
  }
} /* sdlDestroyTriangleList */



/* * *     EDGES: TRIANGLE ADJACENCIES AND PATHS     * * */

void sdlBuildAdjacencies(sdMeshPtr m, sdBoolean keepET)
{
  unsigned Nedges, hash;
  int n, i, clash;
  sdEdgePtr *et, eth;
  sdTrianglePtr *mtaux, t;
  sdPointPtr p0, p1, paux;
  sdBoolean found;

  if (!m)
    return;

  /* create a hash table for edges and insert all three edges of each	*/
  /* initial triangle to it -- NB: Euler & Co. say that, for any	*/
  /* 2-manifold mesh of simply connected polygons, "E = V+F-2+2G+B"	*/
  /* must hold, so "E = (V+F)*1.2" should be enough			*/
  Nedges = (m->Npoints + m->NmotherTriangles) * 6 / 5;
  if (!(et = NEW(Nedges, sdEdgePtr)))
    sdiError("sdlBuildAdjacencies: not enough memory for the edge table", -13);
  
  m->Nedges = clash = 0;
  for (n=0, mtaux=m->motherTriangle; n<m->NmotherTriangles; n++, mtaux++) {
    t = *mtaux;
    for (i=0; i<3; i++) {
      p0 = t->vtx[i];
      p1 = t->vtx[sdP1[i]];
      if (p0->n > p1->n) {
	paux = p1;  p1 = p0;  p0 = paux;
      }
      hash = (0x10001*p0->n + 0x10003*p1->n) % Nedges; /* 6553{7,9} are prime */
      found = FALSE;
      while (eth = et[hash]) {
	if (eth->vtx[0] == p0 && eth->vtx[1] == p1) {
	  /* edge found => fill in adjacency information */
	  if (eth->adt[1])
	    m->nonManifold = TRUE;
	  else {
	    t->adt[i] = eth->adt[0];
	    eth->adt[1] = eth->adt[0]->adt[eth->e4t[0]] = t;
	    eth->e4t[1] = i;
	  }
	  found = TRUE;
	  break;
	}
	/* hash clash => look in next bucket */
	clash++;
	hash++;  if (hash == Nedges) hash = 0;
      }
      if (!found) {
	/* edge not found => insert */
	et[hash] = NEW(1, sdEdge);
	et[hash]->vtx[0] = p0;
	et[hash]->vtx[1] = p1;
	et[hash]->adt[0] = t;
	et[hash]->e4t[0] = i;
	m->Nedges++;
      }
    }
  }

  /* find out whether mesh has a boundary or not and clean up */
  m->closed = TRUE;
  if (keepET) {
    if (!(m->edge = NEW(m->Nedges, sdEdge)))
      sdiError("sdlBuildAdjacencies: not enough memory for the edge table",
	       -13);
    if (!(m->targetMeshPath4edge = NEW(m->Nedges, sdVectorLst)))
      sdiError("sdlBuildAdjacencies: not enough memory for the edge paths",
	       -13);
    for (hash=n=0; hash<Nedges; hash++)
      if (et[hash]) {
	if (!et[hash]->adt[1])
	  m->closed = FALSE;
	m->edge[n++] = *et[hash];
      }
  } else
    for (hash=0; hash<Nedges; hash++)
      if (et[hash] && !et[hash]->adt[1]) {
	m->closed = FALSE;
	break;
      }
  DEL(et);
  if (m->nonManifold)
    fprintf(stderr, "  Warning: non-manifold mesh!\n");
  if (m->verbose > 1) {
    fprintf(stderr, "\n  E = %d actual / %d estimated (%d hash clashes)",
	    m->Nedges, Nedges, clash);
    if (!m->nonManifold) {
      if (m->closed)
	fprintf(stderr, "; B = 0; G = %d",
		(m->Nedges - m->Npoints - m->NmotherTriangles) / 2 + 1);
      else
	fprintf(stderr, "; B > 0; G = ?");
    }
    fprintf(stderr, ".\n");
  }
} /* sdlBuildAdjacencies */


static int sdlTglsSurrPtIntByPlane(sdPointPtr pt, sdVectorPtr normal,
				   float dplane, sdTriangleLst *tl)
{
  int n, i;
  sdTrianglePtr t, t_prv;
  float d1, d2;

  if (!pt || !normal || !pt->t || !tl)
    return 0;

  /* cf. "sdlBuildVNL" */
  t = pt->t;
  n = 0;
  do {
    i = IDX_VTX_IN_T(pt, t);
    d1 = sdmDotProd(normal, &t->vtx[sdP1[i]]->pos) - dplane;
    d2 = sdmDotProd(normal, &t->vtx[sdM1[i]]->pos) - dplane;
    if (SGN(d1) != SGN(d2)) {
      sdlAddTriangle2TriangleList(t, tl);
      n++;
    }
    t_prv = t;
  } while ((t = t->adt[i]) && t != pt->t);

  if (!t) {
    sdlDestroyTriangleList(tl);
    t = t_prv;
    n = 0;
    do {
      i = IDX_VTX_IN_T(pt, t);
      d1 = sdmDotProd(normal, &t->vtx[sdP1[i]]->pos) - dplane;
      d2 = sdmDotProd(normal, &t->vtx[sdM1[i]]->pos) - dplane;
      if (SGN(d1) != SGN(d2)) {
	sdlAddTriangle2TriangleList(t, tl);
	n++;
      }
    } while (t = t->adt[sdM1[i]]);
  }

  return n;
} /* sdlTglsSurrPtIntByPlane */


static void sdlDestroyPath(sdVectorLst *vl)
{
  sdVectorLst vlaux;

  if (!vl)
    return;

  while (vlaux = *vl) {
    *vl = (*vl)->next;
    DEL(vlaux);
  }
} /* sdlDestroyPath */


static void sdlAddPoint2Path(sdVectorPtr pt, sdVectorLst *vl)
{
  sdVectorLst vl_prv, vl_crt, vl_new;

  if (!pt || !vl)
    return;

  if (!(vl_new = NEW(1, sdVectorLstNode)))
    sdiError("sdlAddpoint2Path: not enough memory for a new point", -13);
  vl_new->pos = *pt;
  for (vl_prv=NULL, vl_crt=*vl; vl_crt; vl_crt=vl_crt->next)
    vl_prv=vl_crt;
  if (vl_prv)
    vl_prv->next = vl_new;
  else
    *vl = vl_new;
} /* sdlAddPoint2Path */


int sdlFindPathBetweenPtsInsidePlane(sdMeshPtr m, sdPointPtr pa, sdPointPtr pb,
				     sdVectorPtr normal, void *from,
				     sdVectorLst *vl)
{
  float dplane, d[3], k;
  int na, n, i, *npts, minpts, shortest;
  sdTriangleLst tla = NULL, tlb = NULL, tlaaux, tlaprv, tlanxt, tlaauxaux;
  sdVectorLst *path, pathaux;
  sdTrianglePtr t_prev, t, t_next;
  sdPointPtr p1, p2, p1or2;
  sdVector v;
  sdBoolean done;
  static sdPointPtr pa_orig;
  static int nrecalls;

  if (!pa || !pb || !normal || !vl)
    return 0;

  /* if the target has been reached, add it to the path and be happy */
  if (pa == pb) {
    sdlAddPoint2Path(&pa->pos, vl);
    return 1;
  }

  /* get outtahere if this path is leading us nowhere */
  if (!from) {
    pa_orig = pa;
    nrecalls = 0;
  } else {
    if (pa == pa_orig) {
      fprintf(stderr, "\n  Warning: closed loop detected in sdlFPBPIP!  ");
      return 0;
    }
    if (++nrecalls >= m->NmaxNestedCalls2FPBPIP) {
      fprintf(stderr, "\n  Warning: %d nested calls to sdlFPBPIP!  ",
	      m->NmaxNestedCalls2FPBPIP);
      return 0;
    }
  }

  /* collect triangles surrounding "pa" and eventually discard some if	*/
  /* this is a recursive call ("from" will point to an already visited	*/
  /* triangle or vertex)						*/
  dplane = sdmDotProd(normal, &pa->pos);
  na = sdlTglsSurrPtIntByPlane(pa, normal, dplane, &tla);
  if (from)
    for (tlaprv=NULL, tlaaux=tla; tlaaux; tlaaux=tlanxt) {
      t = tlaaux->t;
      tlanxt = tlaaux->next;
      if (t == (sdTrianglePtr) from || t->vtx[0] == (sdPointPtr) from ||
	  t->vtx[1] == (sdPointPtr) from || t->vtx[2] == (sdPointPtr) from) {
	if (tlaprv)
	  tlaprv->next = tlanxt;
	else
	  tla = tlanxt;
	DEL(tlaaux);  na--;
      } else
	tlaprv = tlaaux;
    }

  /* try all remaining outgoing paths and find shortest */
  if (!(path = NEW(na, sdVectorLst)) || !(npts = NEW(na, int)))
    sdiError("sdlFindPathBetweenPtsInsidePlane: not enough memory for a new "
	     "set of paths", -13);
  sdlTglsSurrPtIntByPlane(pb, normal, dplane, &tlb);
  minpts = shortest = -1;
  for (tlaaux=tla, n=0; tlaaux; tlaaux=tlaaux->next, n++) {
    sdlAddPoint2Path(&pa->pos, path+n);  npts[n]++;
    t = tlaaux->t;
    i = IDX_VTX_IN_T(pa, t);
    p1 = t->vtx[sdP1[i]];  d[1] = sdmDotProd(normal, &p1->pos) - dplane;
    p2 = t->vtx[sdM1[i]];  d[2] = sdmDotProd(normal, &p2->pos) - dplane;
    if (NEGLIGIBLE(d[1]))
      p1or2 = p1;
    else if (NEGLIGIBLE(d[2]))
      p1or2 = p2;
    else
      p1or2 = NULL;
    if (p1or2) {
      npts[n] += sdlFindPathBetweenPtsInsidePlane(m, p1or2, pb, normal, pa,
						  path+n);
      /* remove from the list of outgoing triangles the one that shares	*/
      /* the edge "pa-p1or2" with "t"					*/
      for (tlaprv=tlaaux, tlaauxaux=tlaaux->next; tlaauxaux;
	   tlaprv=tlaauxaux, tlaauxaux=tlaauxaux->next)
	if (IDX_VTX_IN_T(p1or2, tlaauxaux->t) >= 0) {
	  tlaprv->next = tlaauxaux->next;
	  DEL(tlaauxaux);  na--;
	  break;
	}
      done = TRUE;
    } else {
      t_next = t->adt[sdP1[i]];
      v = p2->pos;  DEC_VEC(v, p1->pos);
      k = d[1] / (d[1] - d[2]);  MUL_VEC(v, k);
      INC_VEC(v, p1->pos);
      sdlAddPoint2Path(&v, path+n);  npts[n]++;
      t_prev = t;
      done = FALSE;
    }
    while (!done && !sdlIsTriangleInTriangleList(t = t_next, tlb)) {
      for (i=0; !done && i<3; i++) {
	d[i] = sdmDotProd(normal, &t->vtx[i]->pos) - dplane;
	if (NEGLIGIBLE(d[i])) {
	  npts[n] += sdlFindPathBetweenPtsInsidePlane(m, t->vtx[i], pb, normal,
						      t, path+n);
	  done = TRUE;
	}
      }
      if (done)
	break;
      if (SGN(d[0]) == SGN(d[1]))
	t_next = t->adt[i = (t->adt[1] == t_prev? 2 : 1)];
      else if (SGN(d[1]) == SGN(d[2]))
	t_next = t->adt[i = (t->adt[2] == t_prev? 0 : 2)];
      else if (SGN(d[2]) == SGN(d[0]))
	t_next = t->adt[i = (t->adt[0] == t_prev? 1 : 0)];
      else
	sdiError("sdlFindPathBetweenPtsInsidePlane: yuyu", -8);
      v = t->vtx[sdP1[i]]->pos;  DEC_VEC(v, t->vtx[i]->pos);
      k = d[i] / (d[i] - d[sdP1[i]]);  MUL_VEC(v, k);
      INC_VEC(v, t->vtx[i]->pos);
      sdlAddPoint2Path(&v, path+n);  npts[n]++;
      t_prev = t;
    }
    if (!done) {
      sdlAddPoint2Path(&pb->pos, path+n);  npts[n]++;
    }
    if (minpts < 0 || npts[n] < minpts)
      minpts = npts[shortest = n];
  }

  /* copy shortest path to "vl", clean up and return shortest path length */
  for (pathaux=path[shortest]; pathaux; pathaux=pathaux->next)
    sdlAddPoint2Path(&pathaux->pos, vl);
  for (n=0; n<na; n++)
    sdlDestroyPath(path+n);
  DEL(npts);
  DEL(path);
  sdlDestroyTriangleList(&tlb);
  sdlDestroyTriangleList(&tla);
  return minpts;
} /* sdlFindPathBetweenPtsInsidePlane */



/* * *    MESHES    * * */

sdMeshPtr sdNewMesh(void)
{
  sdMeshPtr m;

  if (!(m = NEW(1, sdMesh)))
    sdiError("sdNewMesh: not enough memory for a new mesh", -13);
  return m;
} /* sdNewMesh */


void sdDestroyMesh(sdMeshPtr *m)
{
  sdMeshPtr mm;
  int i;

  if (!m || !(mm = *m))
    return;

  if (mm->verbose)
    fprintf(stderr, "Destroying mesh(es)...  ");
  if (mm->progressive) {
#ifdef DIME
    if (mm->useDiMePlugin)
      dimeDestroy();
#endif
    if (mm->targetMesh) {
      mm->targetMesh->verbose = FALSE;
      sdDestroyMesh(&mm->targetMesh);
    }
    sdlDestroyAllDetailTrees(mm);
    if (mm->fDetails)
      fclose(mm->fDetails);
    DEL(mm->edge);
    for (i=0; i<mm->Nedges; i++)
      sdlDestroyPath(&mm->targetMeshPath4edge[i]);
    DEL(mm->targetMeshPath4edge);
  }
  sdlDestroyOctree(mm, &mm->vertexOT, FALSE);
  sdlDestroyVertexList(mm);
  sdlDestroyAllTriangles(mm);
  sdlDestroyTriangleList(&mm->tmpTriangleList);
  if (mm->verbose)
    fprintf(stderr, "Done.\n");
  DEL(*m);
} /* sdDestroyMesh */

